class SweepingViewException(Exception):
    pass


class InvalidReplayError(SweepingViewException):
    def __init__(self, replay, message=None):
        super().__init__(
            "The replay file {} is invalid: {}".format(
                replay, message if message else "No details supplied"
            )
        )


class UnknownFormatVersionError(SweepingViewException):
    def __init__(self, replay, version):
        super().__init__(
            "The replay file {} is using a format version this library doesn't understand: {}".format(
                replay,
                version,
            )
        )


class MimeTypeNotImplemented(SweepingViewException):
    def __init__(self, mime_type):
        super().__init__(
            "The mime type {} is not supported by this library. Sorry!".format(
                mime_type,
            )
        )


class UnknownMimeType(SweepingViewException):
    def __init__(self, mime_type):
        from .mime_types import supported

        super().__init__(
            "Unknown mime type {}!\nAvailable mime types:\n".format(
                mime_type,
                "\n".join(supported()),
            )
        )




from sweeping_view.rmv import RMVReplay

def test_rmv(replay_path):
    rmv = RMVReplay.from_file(replay_path / "test_subject.rmv")

    assert rmv.player_data["name"] == "tkolar"
    assert rmv.properties["level"] == "beginner"
    assert set(rmv.mines) == {
        (1, 3),
        (2, 3),
        (4, 2),
        (4, 3),
        (4, 4),
        (4, 5),
        (4, 6),
        (5, 4),
        (6, 6),
        (7, 6),
    }



# -*- coding: utf-8 -*-

from io import SEEK_CUR
from itertools import count

from .base import BaseReplay
from .exceptions import InvalidReplayError


class AVFReplay(BaseReplay):
    MOUSE_EVENT_TYPES = {
        1: "move",
        3: "lmb_down",
        5: "lmb_up",
        9: "rmb_down",
        17: "rmb_up",
        33: "mmb_down",
        65: "mmb_up",
        145: "rmb_up",
        193: "mmb_up",
        11: "shift_lmb_down",
        21: "lmb_up",
    }

    LEVELS = {
        3: "beginner",
        4: "intermediate",
        5: "expert",
        6: "custom",
    }

    def process_buffer(self, data):
        self.properties = {}
        # version
        self.version = ord(data.read(1))
        self.is_freesweeper = not self.version

        # no idea what these bytes do
        data.read(4)
        level = ord(data.read(1))
        try:
            self.properties["level"] = self.LEVELS[level]
        except KeyError:
            raise InvalidReplayError(self, message="Invalid level!")

        if level == 3:
            self.cols = 8
            self.rows = 8
            self.num_mines = 10
        elif level == 4:
            self.cols = 16
            self.rows = 16
            self.num_mines = 40
        elif level == 5:
            self.cols = 30
            self.rows = 16
            self.num_mines = 99
        elif level == 6:
            self.cols = ord(data.read(1)) + 1
            self.rows = ord(data.read(1)) + 1
            self.num_mines = self.read_int(data.read(2))

        self.mines = []
        for ii in range(self.num_mines):
            row = ord(data.read(1))
            col = ord(data.read(1))
            self.mines.append((row, col))

        while data.read(1) != b"[":
            pass

        data.seek(-3, SEEK_CUR)
        self.properties["questionmarks"] = ord(data.read(1)) == 17

        # read past opening "["
        data.read(2)

        info = b""
        while True:
            char = data.read(1)
            if char == b"]":
                break
            info += char
        # TODO: make sure this is always correct/add encoding param
        # TODO: split this info into bits and make usable
        self.ts_info = info.decode("cp1252")
        ts_fields = self.ts_info.split("|")
        self.bbbv = ts_fields[-1][1:].split("T")[0]

        last = ord(data.read(1))
        while True:
            cur = ord(data.read(1))
            if last == 0 and cur == 1:
                break
            last = cur
        data.seek(-3, SEEK_CUR)

        self.events = []
        while True:
            mouse, x1, s2, x2, hun, y1, s1, y2 = tuple(data.read(8))
            xpos = (x1 << 8) + x2
            ypos = (y1 << 8) + y2
            sec = (s1 << 8) + s2 - 1
            gametime = 1000 * sec + 10 * hun
            self.events.append(
                {
                    "type": "mouse",
                    "subtype": self.MOUSE_EVENT_TYPES[mouse],
                    "gametime": gametime,
                    "xpos": xpos,
                    "ypos": ypos,
                }
            )
            if sec < 0:
                break

        last2, last1 = data.read(2)
        ref = tuple(b"cs=")
        while True:
            cur = ord(data.read(1))
            if cur == b"":
                raise InvalidReplayError(self)
            if (last2, last1, cur) == ref:
                break
            last2, last1 = last1, cur
        if self.is_freesweeper:
            for event in self.events:
                ths = ord(data.read(1)) & 0xF
                event["gametime"] += ths
        data.read(17)
        if self.is_freesweeper:
            while ord(data.read(1)) != 13:
                pass
        footer = data.read()
        footer_fields = footer.split(b"\r")
        footer_meta_info = {}
        footer_positional = []
        for field in footer_fields:
            key, *value = field.split(b":", 2)
            if value:
                (value,) = value
                # TODO: make sure this is always correct/add encoding param
                footer_meta_info[key] = value.decode("cp1252").strip()
            else:
                # TODO: make sure this is always correct/add encoding param
                footer_positional.append(key.decode("cp1252"))
        # the last event will have -1 to signal the end of the events
        # section => extract game time from the second to last event
        self.timeth = self.events[-2]["gametime"]
        self.name, self.version_info = footer_positional


if __name__ == "__main__":
    main()



from .avf import AVFReplay
from .exceptions import MimeTypeNotImplemented, UnknownMimeType
from .rmv import RMVReplay

MIME_TYPES = {
    "application/x-minesweeper-arbiter": AVFReplay,
    "application/x-viennasweeper": RMVReplay,
    "application/x-minesweeper-x": NotImplemented,
}


def get_class(mime_type):
    cls_or_nimpl = MIME_TYPES.get(mime_type, None)
    if cls_or_nimpl is None:
        raise UnknownMimeType(mime_type)
    if cls_or_nimpl is NotImplemented:
        raise MimeTypeNotImplemented(mime_type)
    return cls_or_nimpl


def supported():
    return [mt for mt, cls in MIME_TYPES.items() if cls is not NotImplemented]




import pytest

from sweeping_view.avf import AVFReplay


def test_avf(replay_path):
    avf = AVFReplay.from_file(replay_path / "test_subject.avf")

    assert avf.name == "Tommy"
    assert avf.properties["level"] == "beginner"
    assert set(avf.mines) == {
        (4, 1),
        (5, 2),
        (8, 2),
        (2, 3),
        (6, 3),
        (4, 5),
        (8, 5),
        (6, 6),
        (8, 6),
        (4, 7),
    }



# -*- coding: utf-8 -*-

from io import BytesIO


class BaseReplay:
    def __init__(self, data_buffer, name=None):
        self.name = name
        self.process_buffer(data_buffer)

    def process_buffer(self, data):
        raise NotImplementedError

    def __str__(self):
        return "{}({})".format(type(self).__name__, self.name)

    @classmethod
    def from_file(cls, filename):
        with open(filename, "rb") as file:
            return cls(file, name=filename)

    @classmethod
    def from_bytes(cls, data, name=None):
        return cls(BytesIO(data), name=name)

    @staticmethod
    def read_int(binstr):
        res = 0
        for char in binstr:
            res <<= 8
            res += char
        return res



from setuptools import setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="sweeping-view",
    description="Readers for Minesweeper replays",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Thomas Kolar",
    author_email="thomaskolar90@gmail.com",
    url="https://github.com/ralokt/sweeping-view/",
    packages=["sweeping_view"],
    platforms=["all"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Other Audience",
        "Topic :: Games/Entertainment :: Puzzle Games",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    keywords=[
        "minesweeper",
        "replay",
        "readers",
    ],
)



# -*- coding: utf-8 -*-

import logging

from .base import BaseReplay
from .exceptions import InvalidReplayError, UnknownFormatVersionError

logger = logging.getLogger(__name__)


class RMVReplay(BaseReplay):
    MODES = {
        0: "normal",
        1: "UPK",
        2: "cheat",
        3: "density",
    }

    LEVELS = {
        0: "beginner",
        1: "intermediate",
        2: "expert",
        3: "custom",
    }

    MOUSE_EVENT_TYPES = {
        1: "move",
        2: "lmb_down",
        3: "lmb_up",
        4: "rmb_down",
        5: "rmb_up",
        6: "mmb_down",
        7: "mmb_up",
    }

    BOARD_EVENT_TYPES = {
        9: "pressed",
        10: "pressed_qm",
        11: "closed",
        12: "qm",
        13: "flag",
        14: "open",
        18: "open_0",
        19: "open_1",
        20: "open_2",
        21: "open_3",
        22: "open_4",
        23: "open_5",
        24: "open_6",
        25: "open_7",
        26: "open_8",
        27: "open_blast",
    }

    TERMINATION_EVENT_TYPES = {
        15: "blast",
        16: "win",
        17: "other",
    }

    PLAYER_FIELDS = (
        "name",
        "nickname",
        "country",
        "token",
    )

    def process_buffer(self, data):
        # header 1
        extension = data.read(4)
        ftype = self.read_int(data.read(2))
        filesize = self.read_int(data.read(4))

        if ftype != 1:
            raise UnknownFormatVersionError(self, ftype)

        # header 2
        result_str_size = self.read_int(data.read(2))
        version_info_size = self.read_int(data.read(2))
        player_info_size = self.read_int(data.read(2))
        board_size = self.read_int(data.read(2))
        preflagged_size = self.read_int(data.read(2))
        properties_size = self.read_int(data.read(2))
        vid_size = self.read_int(data.read(4))
        checksum_size = self.read_int(data.read(2))

        # result string
        result_str = data.read(result_str_size)

        # version information
        self.version_info = data.read(version_info_size)

        # player fields
        num_player_fields = self.read_int(data.read(2))
        player_fields = []
        player_data = {}
        for _ in range(num_player_fields):
            field_size = ord(data.read(1))
            player_fields.append(data.read(field_size))

        for field_name, field_value in zip(self.PLAYER_FIELDS, player_fields):
            player_data[field_name] = field_value

        # board
        self.timestamp_boardgen = self.read_int(data.read(4))
        self.cols = ord(data.read(1))
        self.rows = ord(data.read(1))
        self.num_mines = self.read_int(data.read(2))
        self.mines = []
        for _ in range(self.num_mines):
            col = ord(data.read(1))
            row = ord(data.read(1))
            self.mines.append((row, col))

        # preflagged
        self.preflags = []
        if preflagged_size:
            num_preflags = self.read_int(data.read(2))
            for _ in range(num_preflags):
                col = ord(data.read(1))
                row = ord(data.read(1))
                self.preflags.append((row, col))

        # properties
        properties = []
        self.properties = {}
        for _ in range(properties_size):
            properties.append(ord(data.read(1)))

        self.properties["questionmarks"] = bool(properties[0])
        self.properties["nonflagging"] = bool(properties[1])

        try:
            self.properties["mode"] = self.MODES[properties[2]]
        except KeyError:
            raise InvalidReplayError(self, message="Invalid mode!")
        try:
            self.properties["level"] = self.LEVELS[properties[3]]
        except KeyError:
            raise InvalidReplayError(self, message="Invalid level!")
        if properties_size > 4 and properties[4]:
            encoding = "utf-8"
        else:
            # TODO: verify or add encoding parameter
            encoding = "cp1252"
        self.events = []
        self.player_data = {
            name: value.decode(encoding) for name, value in player_data.items()
        }

        result_str_field_list = [
            i.strip() for i in result_str.decode(encoding).split("#") if i.strip()
        ]
        self.result_str_dict = {
            key.strip(): value.strip()
            for key, value in (field.split(":") for field in result_str_field_list)
        }
        self.bbbv = int(self.result_str_dict["3BV"])

        while data:
            evcode = ord(data.read(1))
            if evcode == 0:
                logger.warning("Warning, timestampchange is deprecated!")
                new_timestamp = self.read_int(data.read(4))
                self.events.append(
                    {
                        "type": "timestamp_change",
                        "new_timestamp": new_timestamp,
                    }
                )
            elif 1 <= evcode <= 7:
                gametime = self.read_int(data.read(3))
                numflags = ord(data.read(1))
                xpos = self.read_int(data.read(2))
                ypos = self.read_int(data.read(2))
                self.events.append(
                    {
                        "type": "mouse",
                        "subtype": self.MOUSE_EVENT_TYPES[evcode],
                        "gametime": gametime,
                        "numflags": numflags,
                        # compensate for vsweep offsetting these coords from
                        # the upper left corner of the window, or whatever
                        # (it is not the board, in any case)
                        "xpos": xpos - 12,
                        "ypos": ypos - 56,
                    }
                )

            elif 9 <= evcode <= 14 or 18 <= evcode <= 27:
                col = ord(data.read(1))
                row = ord(data.read(1))
                self.events.append(
                    {
                        "type": "board",
                        "subtype": self.BOARD_EVENT_TYPES[evcode],
                        "col": col,
                        "row": row,
                    }
                )

            elif 15 <= evcode <= 17:
                self.events.append(
                    {
                        "type": "terminate",
                        "how": self.TERMINATION_EVENT_TYPES[evcode],
                    }
                )
                break
            else:
                raise InvalidReplayError(self)

        self.timeth = self.read_int(data.read(3))
        self.checksum = data.read(checksum_size)
